#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os;
from string import Template

class template:
    space_chars = 4 * ' ';

    def __init__(self, config):
        self.config = config;
        tplPath = config.template;
        try:
            tplFile = open(tplPath, 'r')
            self.tplContent = tplFile.read();
            tplFile.close()
        except IOError:
            print("template file open faild, file path:%" % tplPath)

    def write(self):
        tpl = Template(self.tplContent)

        nbproc = 1
        sockets = []
        acl = [];
        use_backend = []
        backend = []
        default_backend = ""

        if( len(self.config.socketFiles) > 0 ):
            nbproc = len(self.config.socketFiles)
            idx = 0
            for socketFile in self.config.socketFiles:
                idx += 1
                sockets.append("\n%sstats socket %s mode 600 level admin process %d" % (template.space_chars, socketFile, idx))

        if( len(self.config.groups) > 0 ):
            for group_name in self.config.groups:
                group = self.config.groups[group_name]
                #default backend
                if( group.defaultBackend ):
                    default_backend = "\n%sdefault_backend %s" % (template.space_chars, group_name)

                 #acl
                if( group.acls and len(group.acls) > 0 ):
                    idx = 0
                    for acl_value in group.acls:
                        idx +=1
                        acl_name = group_name + str(idx)
                        #acl
                        acl.append("\n%sacl %s %s" % (template.space_chars, acl_name, acl_value) )
                        #use_backend
                        use_backend.append("\n%suse_backend %s if %s" % (template.space_chars, group_name, acl_name) )

                #backend
                backend.append("\n\nbackend %s" % group_name)
                backend.append("\n%smode %s" % (template.space_chars, group.mode) )
                backend.append("\n%sbalance %s" % (template.space_chars, group.balance) )
                if group.httpchk:
                    backend.append("\n%soption httpchk %s" % (template.space_chars, group.httpchk) )
                backend.append("\n%sserver-template %s" % (template.space_chars, group.serverTemplate) )
                #for inst in group.instances:
                #    backend.append("\n%sserver %s-%s %s check inter 2000 rise 1 fall 2 " % (template.space_chars, group_name, inst, inst))


        output = tpl.substitute({
            "nbproc": nbproc,
            "pid_file": self.config.pidFile,
            "sockets": ''.join(sockets),
            "acl": ''.join(acl),
            "use_backend": ''.join(use_backend),
            "default_backend": default_backend,
            "backend": ''.join(backend),
            "listener_port": self.config.listenerPort,
            "admin_port": self.config.adminPort
        })

        handle = open(self.config.confFile, 'w')
        handle.write(output)
        handle.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import socket
import sys
import threading
from proxyzoo import utils
from proxyzoo.httpserver.response import response

class server(threading.Thread):
    def __init__(self, config_all):
        threading.Thread.__init__(self)
        self.config_all = config_all
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def __del__(self):
        if( hasattr(self, "server_socket") and self.server_socket):
            self.server_socket.close()

    def run(self):
        utils.log("启动HTTP服务，端口：%d" % self.port)
        self.server_socket.listen(128)
        while True:
            client_socket, client_address = self.server_socket.accept()
            #print("[%s, %s]用户连接上了" % client_address)
            handle_client_process = response(client_socket, self.config_all)
            handle_client_process.start()
            client_socket.close()

    def bind(self, port):
        self.port = port
        self.server_socket.bind(("", port))


if __name__ == "__main__":
    s = server({})
    s.bind(8000)
    s.start()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from proxyzoo import utils

DEFAULT_BE_MODE = "http"    #默认backend的模式为http
DEFAULT_BE_BALANCE = "roundrobin"  #默认backend的均衡策略为roundrobin
DEFAULT_BE_SERVER_TEMPLATE = "${group_name} 1-${max_server} 0.0.0.0:0 check inter 2000 rise 1 fall 2 disabled" #默认backend的server模板
DEFAULT_BE_MAX_SERVER = 10  #默认backend的最大server数量

class config:

    def __init__(self, name, zkAddr, zkPath, exec, template, confFile, pidFile, socketFiles, listenerPort, adminPort):
        self.name = name;
        self.zkAddr = zkAddr;
        self.zkPath = zkPath;
        self.exec = utils.parse_path(exec);
        self.template = utils.parse_path(template)
        self.confFile = utils.parse_path(confFile)
        self.pidFile = utils.parse_path(pidFile)
        self.socketFiles = [];
        if(socketFiles):
            for socketFile in socketFiles:
                self.socketFiles.append(utils.parse_path(socketFile))
        self.listenerPort = listenerPort
        self.adminPort = adminPort
        self.groups = {}

    def __str__(self):
        return "config\n name:%s, zk.addr:%s, zk.path:%s " % (self.name, self.zkAddr, self.zkPath)

    def add_group(self, groupname, group):
        self.groups[groupname] = group

    def get_group(self, groupname):
        return self.groups[groupname] if groupname in self.groups else None

    class group:
        def __init__(self, name, zkPath, acls, mode, balance, httpchk, serverTemplate, maxServer, defaultBackend):
            self.name = name
            self.zkPath = zkPath
            self.acls = acls
            self.mode = mode if mode else DEFAULT_BE_MODE
            self.balance = balance if balance else DEFAULT_BE_BALANCE
            self.httpchk = httpchk
            self.serverTemplate = serverTemplate if serverTemplate else DEFAULT_BE_SERVER_TEMPLATE
            self.maxServer = maxServer if maxServer and isinstance(maxServer, int) else DEFAULT_BE_MAX_SERVER
            self.defaultBackend = defaultBackend
            self.instances = []
            self.serverTemplate = self.serverTemplate.replace("${group_name}", self.name)
            self.serverTemplate = self.serverTemplate.replace("${max_server}", str(self.maxServer))

        def __str__(self):
            return "group\n name:%s, zkPath: %s, acls:%s, mode:%s, balance:%s, httpchk:%s, serverTemplate:%s, defaultBackend:%s, instances:%s " % (self.name, self.zkPath, self.acls, self.mode, self.balance, self.httpchk, self.serverTemplate, self.defaultBackend, self.instances)

        def add_instance(self, instance):
            if(instance and not instance in self.instances):
                self.instances.append(instance)
                return True
            return False

        def remove_instance(self, instance):
            if(instance and instance in self.instances):
                self.instances.remove(instance)
                return True
            return False

        def get_instances(self):
            return self.instances







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json;
from time import strftime, localtime

def log(str):
    print("[" + strftime("%Y-%m-%d %H:%M:%S", localtime()) + "] " + str)

#加载json文件
def load_jsonfile(cfgPath):
    """"读取配置"""
    with open(cfgPath, 'r') as json_file:
        config = json.load(json_file)
    return config

#解析路径，替换HOME变量
def parse_path(path):
    if(not path or not isinstance(path,str)): return None
    path = path.replace("${HOME}", "~")
    path = path.replace("$HOME", "~")
    path = path.replace("/", os.sep)
    path = path.replace("\\", os.sep)
    return os.path.abspath(os.path.expanduser(path))



import os
import time
import threading
from proxyzoo import utils
from proxyzoo.haproxy.executor import executor
from proxyzoo.haproxy.template import template
from kazoo.client import KazooClient,ChildrenWatch,DataWatch
from kazoo.handlers.threading import KazooTimeoutError

class sync(threading.Thread):
    def __init__(self, lock, config):
        threading.Thread.__init__(self)
        self.name = "zksync-thread-%s" % config.name
        self.lock = lock
        self.config = config
        self.executor = executor(config)
        self.template = template(config)

    def __del__(self):
        if( hasattr(self, "client") and self.client):
            self.client.close()

    def run(self):
        utils.log("开启服务实例步线程：" + self.name)
        self.client = KazooClient(hosts=self.config.zkAddr)
        try:
         self.client.start(timeout=10)
        except KazooTimeoutError:
            utils.log("%s-连接ZK服务器[%s]超时" % (self.name, self.config.zkAddr))
            return
        self.zkpathGroup = {}

        #第一次启动清理pidfile,终止原有实例
        self.lock.acquire()
        #self.executor.reset()
        self.template.write()
        self.executor.start()
        #utils.log("等待haproxy启动...")
        #time.sleep(3)
        self.lock.release()

        for group_name in self.config.groups:
            group = self.config.groups[group_name]
            zkpath = group.zkPath if group.zkPath else self.config.zkPath
            zkpath = zkpath.replace("${app}", group_name)
            zkpath = zkpath.replace("${group}", group_name)
            self.zkpathGroup[zkpath] = group_name
            self.client.ensure_path(zkpath)
            self.watcher(zkpath)

        while True:
            time.sleep(5)
            #print('watching......')

    def watcher(self, zkpath):
        # 获取原子节点列表
        _old_node_list = self.client.get_children(zkpath)

        #if(len(self._old_node_list) > 0):
        group_name = self.zkpathGroup[zkpath] if zkpath in self.zkpathGroup else None
        group = self.config.get_group(group_name) if group_name else None
        utils.log('[%s-%s]初始节点列表：%s' % (self.config.name, group_name, _old_node_list))

        if group:
            group._old_node_list = _old_node_list
            for node in _old_node_list:
                group.add_instance(str(node))

            #获取锁，用于线程同步
            self.lock.acquire()
            self.executor.init_group_instances(group)
            self.lock.release()

        try:
            # 为所要监听的节点开启一个子节点监听器
            ChildrenWatch(client=self.client, path=zkpath, func=self._node_change, send_event=True)
        except Exception as e:
            raise

    def _node_change(self, new_node_list, event):
        # 这里的new_node_list是指当前最新的子节点列表
        if not event: return

        has_change = False
        zkpath = event.path
        group_name = self.zkpathGroup[zkpath] if zkpath in self.zkpathGroup else None
        group = self.config.get_group(group_name) if group_name else None

        if not group:
            utils.log("group_name[%s]未找到对应的配置信息" % group_name)
            return;

        # 当前节点列表与上次拿到的节点列表相等，注意不是长度相等，是列表值和长度都要相等
        if new_node_list == group._old_node_list:
            utils.log('[%s-%s]子节点列表未发生变化' % (self.config.name, group_name))
            return

        if len(new_node_list) > len(group._old_node_list):
            for new_node in new_node_list:
                if new_node not in group._old_node_list:
                    utils.log('[%s-%s]监听到一个新的节点：%s' % (self.config.name, group_name, str(new_node)) )
                    group._old_node_list = new_node_list
                    inst = str(new_node)
                    if group and group.add_instance(inst):
                        self.lock.acquire()
                        self.executor.add_instance(group, inst)
                        self.lock.release()
        else:
            for old_node in group._old_node_list:
                if old_node not in new_node_list:
                    utils.log('[%s-%s]监听到一个删除的节点：%s' % (self.config.name, group_name, str(old_node)) )
                    group._old_node_list = new_node_list
                    inst = str(old_node)
                    if group and group.remove_instance(str(old_node)):
                        self.lock.acquire()
                        self.executor.remove_instance(group, inst)
                        self.lock.release()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-i

# description: 一个守护进程的简单包装类, 具备常用的start|stop|restart|status功能, 使用方便
#             需要改造为守护进程的程序只需要重写基类的run函数就可以了
# date: 2015-10-29
# usage: 启动: python daemon.py start
#       关闭: python daemon.py stop
#       状态: python daemon.py status
#       重启: python daemon.py restart
#       查看: ps -axj | grep daemon

import atexit
import os
import signal
import sys
import time

from proxyzoo import utils

class daemon:
    '''
    a generic daemon class.
    usage: subclass the CDaemon class and override the run() method
    stderr  表示错误日志文件绝对路径, 收集启动过程中的错误日志
    verbose 表示将启动运行过程中的异常错误信息打印到终端,便于调试,建议非调试模式下关闭, 默认为1, 表示开启
    save_path 表示守护进程pid文件的绝对路径
    '''

    def __init__(self, save_path, stdin=os.devnull, stdout=os.devnull, stderr=os.devnull, home_dir='.', umask=22, verbose=1):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = save_path  # pid文件绝对路径
        self.home_dir = home_dir
        self.verbose = verbose  # 调试开关
        self.umask = umask
        self.daemon_alive = True

    '''
    编写守护进程的一般步骤步骤：
    （1）创建自己成并被init进程接管：在父进程中执行fork并exit退出；
    （2）创建新进程组和新会话：在子进程中调用setsid函数创建新的会话；
    （3）修改子进程的工作目录：在子进程中调用chdir函数，让根目录 ”/” 成为子进程的工作目录；
    （4）修改子进程umask：在子进程中调用umask函数，设置进程的umask为0；
    （5）在子进程中关闭任何不需要的文件描述符
    
    在子进程中再次fork一个进程，这个进程称为孙子进程，之后子进程退出
    重定向孙子进程的标准输入流、标准输出流、标准错误流到/dev/null
    那么最终的孙子进程就称为守护进程。
    '''

    def daemonize(self):
        try:
            '''
            进程调用fork函数时，操作系统会新建一个子进程，它本质上与父进程完全相同。子进程从父进程继承了多个值的拷贝，比如全局变量和环境变量。
            两个进程唯一的区别就是fork的返回值。child（子）进程接收返回值为0，而父进程接收子进程的pid作为返回值。
            调用fork函数后，两个进程并发执行同一个程序，首先执行的是调用了fork之后的下一行代码。父进程和子进程既并发执行，又相互独立。
            '''
            # 创建子进程，而后父进程退出
            # 为避免挂起控制终端将Daemon放入后台执行。方法是在进程中调用fork使父进程终止，让Daemon在子进程中后台执行。
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as e:
            sys.stderr.write('fork #1 failed: %d (%s)\n' % (e.errno, e.strerror))
            sys.exit(1)

        # 创建新会话，子进程成为新会话的首进程（session leader）
        '''
        setsid()函数可以建立一个对话期。
        
        会话期(session)是一个或多个进程组的集合。
        如果，调用setsid的进程不是一个进程组的组长，此函数创建一个新的会话期。
        (1)此进程变成该对话期的首进程
        (2)此进程变成一个新进程组的组长进程。
        (3)此进程没有控制终端，如果在调用setsid前，该进程有控制终端，那么与该终端的联系被解除。 如果该进程是一个进程组的组长，此函数返回错误。
        (4)为了保证这一点，我们先调用fork()然后exit()，此时只有子进程在运行
        '''
        # 创建新的会话，子进程成为会话的首进程
        # 控制终端，登录会话和进程组通常是从父进程继承下来的。我们的目的就是要摆脱它们，使之不受它们的影响。方法是在创建子进程的基础上，调用setsid()使进程成为会话组长
        os.setsid()

        # 修改子进程的工作目录
        os.chdir(self.home_dir)

        '''
        由于umask会屏蔽权限，所以设定为0，这样可以避免读写文件时碰到权限问题。
        '''
        # 修改子进程umask为0
        os.umask(self.umask)

        '''
        现在，进程已经成为无终端的会话组长。但它可以重新申请打开一个控制终端。可以通过使进程不再成为会话组长来禁止进程重新打开控制终端：
        '''
        try:
            # 创建孙子进程，而后子进程退出
            # 新创建的孙子进程，不是会话组长
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as e:
            sys.stderr.write('fork #2 failed: %d (%s)\n' % (e.errno, e.strerror))
            sys.exit(1)

        '''
        因为是守护进程，本身已经脱离了终端，那么标准输入流、标准输出流、标准错误流就没有什么意义了。
        所以都转向到/dev/null，就是都丢弃的意思。
        '''
        # 重定向孙子进程的标准输入流、标准输出流、标准错误流到/dev/null
        sys.stdout.flush()
        sys.stderr.flush()

        si = open(self.stdin, 'r')
        so = open(self.stdout, 'a+')
        if self.stderr:
            se = open(self.stderr, 'a+', 1)
        else:
            se = so

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        def sig_handler(signum, frame):
            self.daemon_alive = False

        signal.signal(signal.SIGTERM, sig_handler)
        signal.signal(signal.SIGINT, sig_handler)

        if self.verbose >= 1:
            print('daemon process started ...')

        atexit.register(self.del_pid)
        pid = str(os.getpid())
        open(self.pidfile, 'w+').write('%s\n' % pid)

    def get_pid(self):
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        except SystemExit:
            pid = None
        return pid

    def del_pid(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def start(self, *args, **kwargs):
        if self.verbose >= 1:
            print('ready to starting ......')
        # check for a pid file to see if the daemon already runs
        pid = self.get_pid()
        if pid:
            msg = 'pid file %s already exists, is it already running?\n'
            sys.stderr.write(msg % self.pidfile)
            sys.exit(1)
        # start the daemon
        self.daemonize()
        self.run(*args, **kwargs)

    def stop(self):
        if self.verbose >= 1:
            print('stopping ...')
        pid = self.get_pid()
        if not pid:
            msg = 'pid file [%s] does not exist. Not running?\n' % self.pidfile
            sys.stderr.write(msg)
            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)
            return
        # try to kill the daemon process
        try:
            i = 0
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
                i = i + 1
                if i % 10 == 0:
                    os.kill(pid, signal.SIGHUP)
        except OSError as err:
            err = str(err)
            if err.find('No such process') > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err))
                sys.exit(1)
            if self.verbose >= 1:
                print('Stopped!')

    def restart(self, *args, **kwargs):
        self.stop()
        self.start(*args, **kwargs)

    def is_running(self):
        pid = self.get_pid()
        # print(pid)
        return pid and os.path.exists('/proc/%d' % pid)

    def run(self, *args, **kwargs):
        'NOTE: override the method in subclass'
        print('base class run()')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import threading
from proxyzoo import utils
from proxyzoo.conf.config import config
from proxyzoo.zk.sync import sync
from proxyzoo.httpserver.server import server

def start(configPath, httpServerPort=21000):
    cfgPath = utils.parse_path(configPath)
    cfgJson = utils.load_jsonfile(cfgPath)

    #判断并设置默认服务端口
    httpServerPort = httpServerPort if isinstance(httpServerPort, int) else 21000

    lock = threading.Lock()
    threads = []
    config_all = {}
    for name in cfgJson:
        item = cfgJson[name];
        cfg = config(name,
                    item['zk.addr'],
                    item['zk.path'],
                    item['exec'],
                    item['template'],
                    item['conf_file'],
                    item['pid_file'],
                    item['socket_files'],
                    item['listener_port'],
                    item['admin_port'],
               )
        for group_name in item["groups"]:
            group_item = item["groups"][group_name]
            zkPath = group_item['zk.path'] if 'zk.path' in group_item else None
            acls = group_item['acls'] if 'acls' in group_item else None
            mode = group_item['mode'] if 'mode' in group_item else None
            balance = group_item['balance'] if 'balance' in group_item else None
            httpchk = group_item['httpchk'] if 'httpchk' in group_item else None
            serverTemplate = group_item['server-template'] if 'server-template' in group_item else None
            maxServer = group_item['max_server'] if 'max_server' in group_item else None
            defaultBackend = True.__eq__(group_item['default_backend']) if 'default_backend' in group_item else False
            cfg_group = config.group(group_name, zkPath, acls, mode, balance, httpchk, serverTemplate, maxServer, defaultBackend)
            #print(cfg_group)
            cfg.add_group(group_name, cfg_group)

        config_all[name] = cfg

        t = sync(lock, cfg)
        t.start()
        threads.append(t)

    #启动HTTP服务
    s = server(config_all)
    s.bind(httpServerPort)
    s.start()

    threads.append(s)

    for t in threads:
        t.join()

if __name__ == '__main__':
    start('./conf/config.json')


import sys
from setuptools import find_packages,setup

install_requires = ['kazoo']

with open("README.rst", "r") as fp:
    proxyzoo_long_description = fp.read()

#text/x-rst
#text/markdown
#text/plain

setup(
    name = 'proxyzoo',
    version = '1.0.2',
    url = "http://git.wadecn.com:18082/bits/proxy-zk",
    author = "xiedx",
    author_email = "40317303@qq.com",
    description = "Haproxy dynamic load balancing",
    long_description = proxyzoo_long_description,
    long_description_content_type = "text/plain",
    license = "MIT",
    classifiers = [
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Networking",
    ],
    #py_modules = ['proxyzoo'],
    #package_dir={'': 'proxyzoo'},
    #packages = ['conf', 'haproxy', 'zk'],
    #packages = ['proxyzoo','proxyzoo.conf', 'proxyzoo.haproxy', 'proxyzoo.zk'],
    packages = find_packages(),
    install_requires = install_requires
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import re
from multiprocessing import Process
from proxyzoo import utils

class response(Process):
    def __init__(self, client_socket, config_all):
        Process.__init__(self)
        self.client_socket = client_socket
        self.config_all = config_all

    def run(self):
        """
        处理客户端请求
        """
        client_socket = self.client_socket

        # 获取客户端请求数据
        request_data = client_socket.recv(1024)
        #print("request data:", request_data)
        request_lines = request_data.splitlines()
        #for line in request_lines:
        #    print(line)

        # 解析请求报文
        request_start_line = request_lines[0]
        # 提取用户请求的文件名及请求方法
        file_name = re.match(r"\w+ +(/[^ ]*) ", request_start_line.decode("utf-8")).group(1)
        method = re.match(r"(\w+) +/[^ ]* ", request_start_line.decode("utf-8")).group(1)
        #print("file_name: %s" % file_name)

        # 构造响应数据
        response_start_line = "HTTP/1.1 200 OK\r\n"
        response_headers = "Cache-Control: no-cache, no-store, must-revalidate\r\nPragma: no-cache\r\nExpires: -1\r\nContent-Type:application/json;charset=utf-8\r\n"  #text/plain
        response_body = "{}"

        #使用正则表达式解析URL参数
        re_cfg_name = re.match(r"/([^/]*)", file_name)
        re_group_name = re.match(r"/[^/]*/([^/]*)", file_name)
        #print("re_cfg_name:%s, re_app_name:%s" % (re_cfg_name, re_app_name))
        cfg_name = re_cfg_name.group(1) if re_cfg_name else None
        group_name = re_group_name.group(1) if re_group_name else None
        utils.log("cfg_name:%s group_name:%s" % (cfg_name, group_name))

        if( "/" == file_name ):
            response_body = self.build_all()
        elif( cfg_name and cfg_name in self.config_all ):
            if( group_name and group_name in self.config_all[cfg_name].groups ):
                response_body = self.build_group(cfg_name, group_name)
            else:
                response_body = self.build_cfg(cfg_name)
        else:
            response_start_line = "HTTP/1.1 403 Forbidden\r\n"

        response = response_start_line + response_headers + "\r\n" + response_body
        #向客户端返回响应数据
        client_socket.send(bytes(response, "utf-8"))
        #关闭客户端连接
        client_socket.close()

    def build_all(self):
        str = "{"
        l = len(self.config_all); idx = 0
        for cfg_name in self.config_all:
            str += "\"" + cfg_name + "\":"
            str += self.build_cfg_body(cfg_name)
            idx += 1
            if idx < l: str += ","
        str += "}"
        return str

    def build_cfg(self, cfg_name):
        str = "{"
        if cfg_name in self.config_all:
            str += "\"" + cfg_name + "\":"
            str += self.build_cfg_body(cfg_name)
        str += "}"
        return str

    def build_group(self, cfg_name, group_name):
        str = "{"
        if cfg_name in self.config_all:
            str += "\"" + cfg_name + "\":"
            str += "{"
            if group_name in self.config_all[cfg_name].groups:
                str += "\"" + group_name + "\":"
                str += self.build_group_body(cfg_name, group_name)
            str +="}"
        str += "}"
        return str

    def build_cfg_body(self, cfg_name):
         str = ""
         if( cfg_name in self.config_all ):
            cfg = self.config_all[cfg_name]
            str = "{"
            l = len(cfg.groups); idx = 0
            for group_name in cfg.groups:
                str += "\"" + group_name + "\":"
                str += self.build_group_body(cfg_name, group_name)
                idx += 1
                if idx < l: str += ","
            str += "}"
         else:
             str = "{}"
         return str

    def build_group_body(self, cfg_name, group_name):
        str = ""
        if( cfg_name in self.config_all and group_name in self.config_all[cfg_name].groups):
            group = self.config_all[cfg_name].groups[group_name]
            str = "{"
            str += "\"instances:\"["
            l = len(group.instances); idx = 0
            for inst in group.instances:
                str += "\"" + inst + "\""
                idx += 1
                if idx < l: str += ","
            str += "]}"
        else:
            str = "{}"
        return str


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import socket
from proxyzoo import utils

'''
---------------------------------------------------------------------
 haproxy backend server 状态属性：
 srv_op_state:                Server operational state (UP/DOWN/...).
                                0 = SRV_ST_STOPPED
                                  The server is down.
                                1 = SRV_ST_STARTING
                                  The server is warming up (up but
                                  throttled).
                                2 = SRV_ST_RUNNING
                                  The server is fully up.
                                3 = SRV_ST_STOPPING
                                  The server is up but soft-stopping
                                  (eg: 404).
 srv_admin_state:             Server administrative state (MAINT/DRAIN/...).
                              The state is actually a mask of values :
                                0x01 = SRV_ADMF_FMAINT
                                  The server was explicitly forced into
                                  maintenance.
                                0x02 = SRV_ADMF_IMAINT
                                  The server has inherited the maintenance
                                  status from a tracked server.
                                0x04 = SRV_ADMF_CMAINT
                                  The server is in maintenance because of
                                  the configuration.
                                0x08 = SRV_ADMF_FDRAIN
                                  The server was explicitly forced into
                                  drain state.
                                0x10 = SRV_ADMF_IDRAIN
                                  The server has inherited the drain status
                                  from a tracked server.
                                0x20 = SRV_ADMF_RMAINT
                                  The server is in maintenance because of an
                                  IP address resolution failure.
                                0x40 = SRV_ADMF_HMAINT
                                  The server FQDN was set from stats socket.
-----------------------------------------------------------------------------
 haproxy backend server 设置状态：
set server <backend>/<server> state [ ready | drain | maint ]
  Force a server's administrative state to a new state. This can be useful to
  disable load balancing and/or any traffic to a server. Setting the state to
  "ready" puts the server in normal mode, and the command is the equivalent of
  the "enable server" command. Setting the state to "maint" disables any traffic
  to the server as well as any health checks. This is the equivalent of the
  "disable server" command. Setting the mode to "drain" only removes the server
  from load balancing but still allows it to be checked and to accept new
  persistent connections. Changes are propagated to tracking servers if any.                                  
'''
class executor:

    def __init__(self, config):
        self.config = config

    def reset(self):
        pidFile = self.config.pidFile
        if( os.path.isfile(pidFile) ):
            command = "kill -9 `cat %s` && rm -fr %s" % (pidFile, pidFile)
            os.system(command)

    def start(self):
        exec = self.config.exec;
        confFile = self.config.confFile
        pidFile = self.config.pidFile

        if( not os.path.exists(exec) ):
            print("exec %s is not exists!" % exec)
            return
        if( not os.path.isfile(confFile) ):
            print("conf file %s is not a exists file!" % exec)
            return

        command = "%s" % exec
        if( os.path.isfile(pidFile) ):
            command += " -sf `cat %s`" % pidFile
        command += " -f  %s" % confFile
        #print("command:%s", command)
        os.system(command)

    def send_command_all(self, command):
        result = []
        for socketFile in self.config.socketFiles:
            result.append(self.send_command(socketFile, command))
        return result

    def send_command(self, socket_file, command):
        haproxy_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        #haproxy_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  #TCP模式
        haproxy_sock.settimeout(10)
        retval = ""
        try:
            haproxy_sock.connect(socket_file)
            haproxy_sock.send(bytes(command, "utf-8"))
            while True:
                buf = haproxy_sock.recv(16)
                if buf:
                    retval += str(buf, encoding = "utf-8")
                else:
                    break
            haproxy_sock.close()
        #except:
        #    retval = ""
        finally:
            haproxy_sock.close()
        return retval

    def init_group_instances(self, group):
        group_name = group.name
        utils.log("初始化group：%s" % group_name)
        idx = 0
        for inst in group.instances:
            idx += 1
            inst_values = inst.split(':')
            server_name =  group_name + str(idx)
            utils.log("激活 %s/%s addr %s port %s" % (group_name, server_name, inst_values[0], inst_values[1]))
            command = "set server %s/%s addr %s port %s\n" % (group_name, server_name, inst_values[0], inst_values[1])
            result = self.send_command_all(command)
            #print(result)
            command = "set server %s/%s state ready\n" % (group_name, server_name)
            result = self.send_command_all(command)
            #print(result)


    def add_instance(self, group, inst):
        # echo "help" | socat stdio /var/lib/haproxy/haproxy.socket
        # echo "show info" | socat stdio /var/lib/haproxy/haproxy.socket
        # set server xxx state ready|maint
        group_name = group.name
        inst_values = inst.split(':')
        for socketFile in self.config.socketFiles:
            servers_slots = self.send_command(socketFile, "show servers state %s\n" % group_name)
            if not servers_slots:
                utils.log("Failed to get %s servers state from socket: %s" % (group_name, socketFile))
                break;
            server_name = ""
            servers_slots = servers_slots.split('\n')
            for server_slot in servers_slots:
                server_infos = server_slot.split(" ")
                if len(server_infos) > 18 and not server_infos[0] == '#':
                    server_op_state = server_infos[5]
                    #server_admin_state = server_infos[6]
                    #server_ip = server_infos[4]
                    #server_port = server_infos[18]
                    if( "0".__eq__(server_op_state) ):
                        server_name = server_infos[3]
                        break
            if( server_name and not "" == server_name ):
                 utils.log("激活 %s/%s addr %s port %s 在%s的进程上" % (group_name, server_name, inst_values[0], inst_values[1], socketFile) )
                 command = "set server %s/%s addr %s port %s\n" % (group_name, server_name, inst_values[0], inst_values[1])
                 result = self.send_command(socketFile, command)
                 #print(result)
                 command = "set server %s/%s state ready\n" % (group_name, server_name)
                 result = self.send_command(socketFile, command)
                 #print(result)
            else:
                utils.log("%s可用的server数量不足: %s" % (group_name, socketFile))

    def remove_instance(self, group, inst):
        # echo "help" | socat stdio /var/lib/haproxy/haproxy.socket
        # echo "show info" | socat stdio /var/lib/haproxy/haproxy.socket
        # set server xxx state ready|maint
        group_name = group.name
        inst_values = inst.split(':')
        for socketFile in self.config.socketFiles:
            servers_slots = self.send_command(socketFile, "show servers state %s\n" % group_name)
            if not servers_slots:
                utils.log("Failed to get %s servers state from socket: %s" % (group_name, socketFile))
                break;
            target_servers = []
            servers_slots = servers_slots.split('\n')
            for server_slot in servers_slots:
                server_infos = server_slot.split(" ")
                if len(server_infos) > 18 and not server_infos[0] == '#':
                    server_name = server_infos[3]
                    #server_op_state = server_infos[5]
                    #server_admin_state = server_infos[6]
                    server_ip = server_infos[4]
                    server_port = server_infos[18]
                    if( server_ip == inst_values[0] and  server_port == inst_values[1] ):
                        target_servers.append(server_name)

            for server_name in target_servers:
                utils.log("下线 %s/%s addr %s port %s 在%s的进程上" % (group_name, server_name, inst_values[0], inst_values[1], socketFile) )
                command = "set server %s/%s state maint\n" % (group_name, server_name)
                result = self.send_command(socketFile, command)
                #print(result)